from o3seespy.command.element.base_element import ElementBase


class ElastomericBearingPlasticity2D(ElementBase):
    """
    The ElastomericBearingPlasticity2D Element Class
    
    This command is used to construct an elastomericBearing element object, which is defined by two nodes. The element
    can have zero length or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D)
    plasticity properties for the shear deformations, and force-deformation behaviors defined by
    UniaxialMaterials in the remaining two (2D) or four (3D) directions. By default (sDratio =
    0.5) P-Delta moments are equally distributed to the two end-nodes. To avoid the
    introduction of artificial viscous damping in the isolation system (sometimes
    referred to as "damping leakage in the isolation system"), the bearing
    element does not contribute to the Rayleigh damping by default. If
    the element has non-zero length, the local x-axis is determined
    from the nodal geometry unless the optional x-axis vector is
    specified in which case the nodal geometry is ignored and
    the user-defined orientation is utilized.

    For a two-dimensional problem
    """
    op_type = 'elastomericBearingPlasticity'

    def __init__(self, osi, ele_nodes, k_init, qd, alpha1, alpha2, mu, p_mat=None, mz_mat=None, do_rayleigh=False, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for ElastomericBearingPlasticity2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        k_init: float
            Initial elastic stiffness in local shear direction
        qd: float
            Characteristic strength
        alpha1: float
            Post yield stiffness ratio of linear hardening component
        alpha2: float
            Post yield stiffness ratio of non-linear hardening component
        mu: float
            Exponent of non-linear hardening component
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.5)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> o3.element.ElastomericBearingPlasticity2D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
        >>>                                           mu=1.0, p_mat=p_mat, mz_mat=mz_mat)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.k_init = float(k_init)
        self.qd = float(qd)
        self.alpha1 = float(alpha1)
        self.alpha2 = float(alpha2)
        self.mu = float(mu)
        self.p_mat = p_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.k_init, self.qd, self.alpha1, self.alpha2, self.mu]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class ElastomericBearingPlasticity3D(ElementBase):
    """
    The ElastomericBearingPlasticity3D Element Class
    
    This command is used to construct an elastomericBearing element object, which is defined by two nodes. The element
    can have zero length or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D)
    plasticity properties for the shear deformations, and force-deformation behaviors defined by
    UniaxialMaterials in the remaining two (2D) or four (3D) directions. By default (sDratio =
    0.5) P-Delta moments are equally distributed to the two end-nodes. To avoid the
    introduction of artificial viscous damping in the isolation system (sometimes
    referred to as "damping leakage in the isolation system"), the bearing
    element does not contribute to the Rayleigh damping by default. If
    the element has non-zero length, the local x-axis is determined
    from the nodal geometry unless the optional x-axis vector is
    specified in which case the nodal geometry is ignored and
    the user-defined orientation is utilized.

    For a three-dimensional problem
    """
    op_type = 'elastomericBearingPlasticity'

    def __init__(self, osi, ele_nodes, k_init, qd, alpha1, alpha2, mu, p_mat=None, t_mat=None, my_mat=None, mz_mat=None, do_rayleigh=False, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for ElastomericBearingPlasticity3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        k_init: float
            Initial elastic stiffness in local shear direction
        qd: float
            Characteristic strength
        alpha1: float
            Post yield stiffness ratio of linear hardening component
        alpha2: float
            Post yield stiffness ratio of non-linear hardening component
        mu: float
            Exponent of non-linear hardening component
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        t_mat: obj, optional
            Object associated with previously-defined uniaxial_material in torsional direction
        my_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local y-axis
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.5)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [0, 1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> orient_vals = [1, 0, 0]
        >>> o3.element.ElastomericBearingPlasticity3D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
        >>>                                           mu=1.0, p_mat=p_mat, t_mat=t_mat, my_mat=my_mat, mz_mat=mz_mat,
        >>>                                           orient=orient_vals)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.k_init = float(k_init)
        self.qd = float(qd)
        self.alpha1 = float(alpha1)
        self.alpha2 = float(alpha2)
        self.mu = float(mu)
        self.p_mat = p_mat
        self.t_mat = t_mat
        self.my_mat = my_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.k_init, self.qd, self.alpha1, self.alpha2, self.mu]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 't_mat') is not None:
            self._parameters += ['-T', self.t_mat.tag]
        if getattr(self, 'my_mat') is not None:
            self._parameters += ['-My', self.my_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class ElastomericBearingBoucWen2D(ElementBase):
    """
    The ElastomericBearingBoucWen2D Element Class

    This command is used to construct an elastomericBearing element object, which is defined by two nodes. The element
    can have zero length or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D)
    plasticity properties for the shear deformations, and force-deformation behaviors defined by
    UniaxialMaterials in the remaining two (2D) or four (3D) directions. By default (sDratio =
    0.5) P-Delta moments are equally distributed to the two end-nodes. To avoid the
    introduction of artificial viscous damping in the isolation system (sometimes
    referred to as "damping leakage in the isolation system"), the bearing
    element does not contribute to the Rayleigh damping by default. If
    the element has non-zero length, the local x-axis is determined
    from the nodal geometry unless the optional x-axis vector is
    specified in which case the nodal geometry is ignored and
    the user-defined orientation is utilized.

    For a two-dimensional problem
    """
    op_type = 'elastomericBearingBoucWen'

    def __init__(self, osi, ele_nodes, k_init, qd, alpha1, alpha2, mu, eta, beta, gamma, p_mat=None, mz_mat=None,
                 orient_vals: list = None, shear_dist: float = None, do_rayleigh=False, mass: float = None):
        """
        Initial method for ElastomericBearingBoucWen2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        k_init: float
            Initial elastic stiffness in local shear direction
        qd: float
            Characteristic strength
        alpha1: float
            Post yield stiffness ratio of linear hardening component
        alpha2: float
            Post yield stiffness ratio of non-linear hardening component
        mu: float
            Exponent of non-linear hardening component
        eta: float
            Yielding exponent (sharpness of hysteresis loop corners) (default = 1.0)
        beta: float
            First hysteretic shape parameter (default = 0.5)
        gamma: float
            Second hysteretic shape parameter (default = 0.5)
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        orient_vals: list, optional
            Vector components in global coordinates defining local x-axis , vector components in global coordinates
            defining local y-axis
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.5)
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        mass: float, optional
            Element mass (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> orient_vals = [1, 0, 0, 1, 0, 1]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> o3.element.ElastomericBearingBoucWen2D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
        >>>                                        mu=1.0, eta=1.0, beta=1.0, gamma=1.0, p_mat=p_mat, mz_mat=mz_mat,
        >>>                                        orient_vals=orient_vals, shear_dist=1.0, do_rayleigh=False, mass=1.0)
        """
        self.osi = osi
        self.ele_nodes = [x.tag for x in ele_nodes]
        self.k_init = float(k_init)
        self.qd = float(qd)
        self.alpha1 = float(alpha1)
        self.alpha2 = float(alpha2)
        self.mu = float(mu)
        self.eta = float(eta)
        self.beta = float(beta)
        self.gamma = float(gamma)
        self.p_mat = p_mat
        self.mz_mat = mz_mat
        self.orient_vals = orient_vals
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        self.do_rayleigh = do_rayleigh
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_nodes, self.k_init, self.qd, self.alpha1, self.alpha2,
                            self.mu, self.eta, self.beta, self.gamma]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'orient_vals') is not None:
            self._parameters += ['-orient', *self.orient_vals]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        try:
            self.to_process(osi)
        except ValueError:
            self._parameters[0] = 'ElastomericBearingBoucWen'
            self.to_process(osi)



class ElastomericBearingBoucWen3D(ElementBase):
    """
    The ElastomericBearingBoucWen3D Element Class

    This command is used to construct an elastomericBearing element object, which is defined by two nodes. The element
    can have zero length or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D)
    plasticity properties for the shear deformations, and force-deformation behaviors defined by
    UniaxialMaterials in the remaining two (2D) or four (3D) directions. By default (sDratio =
    0.5) P-Delta moments are equally distributed to the two end-nodes. To avoid the
    introduction of artificial viscous damping in the isolation system (sometimes
    referred to as "damping leakage in the isolation system"), the bearing
    element does not contribute to the Rayleigh damping by default. If
    the element has non-zero length, the local x-axis is determined
    from the nodal geometry unless the optional x-axis vector is
    specified in which case the nodal geometry is ignored and
    the user-defined orientation is utilized.

    For a three-dimensional problem
    """
    op_type = 'elastomericBearingBoucWen'

    def __init__(self, osi, ele_nodes, k_init, qd, alpha1, alpha2, mu, eta, beta, gamma, p_mat=None, t_mat=None,
                 my_mat=None, mz_mat=None, orient_vals: list = None, shear_dist: float = None, do_rayleigh=False,
                 mass: float = None):
        """
        Initial method for ElastomericBearingBoucWen3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        k_init: float
            Initial elastic stiffness in local shear direction
        qd: float
            Characteristic strength
        alpha1: float
            Post yield stiffness ratio of linear hardening component
        alpha2: float
            Post yield stiffness ratio of non-linear hardening component
        mu: float
            Exponent of non-linear hardening component
        eta: float
            Yielding exponent (sharpness of hysteresis loop corners) (default = 1.0)
        beta: float
            First hysteretic shape parameter (default = 0.5)
        gamma: float
            Second hysteretic shape parameter (default = 0.5)
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        t_mat: obj, optional
            Object associated with previously-defined uniaxial_material in torsional direction
        my_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local y-axis
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        orient_vals: list, optional
            Vector components in global coordinates defining local x-axis , vector components in global coordinates
            defining local y-axis
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.5)
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        mass: float, optional
            Element mass (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [0, 1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> orient_vals = [1, 0, 0]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> o3.element.ElastomericBearingBoucWen3D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
        >>>                                        mu=1.0, eta=1.0, beta=1.0, gamma=1.0, p_mat=p_mat, t_mat=t_mat,
        >>>                                        my_mat=my_mat, mz_mat=mz_mat, orient_vals=orient_vals,
        >>>                                        shear_dist=1.0, do_rayleigh=False, mass=1.0)
        """
        self.osi = osi
        self.ele_nodes = [x.tag for x in ele_nodes]
        self.k_init = float(k_init)
        self.qd = float(qd)
        self.alpha1 = float(alpha1)
        self.alpha2 = float(alpha2)
        self.mu = float(mu)
        self.eta = float(eta)
        self.beta = float(beta)
        self.gamma = float(gamma)
        self.p_mat = p_mat
        self.t_mat = t_mat
        self.my_mat = my_mat
        self.mz_mat = mz_mat
        self.orient_vals = orient_vals
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        self.do_rayleigh = do_rayleigh
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_nodes, self.k_init, self.qd, self.alpha1, self.alpha2,
                            self.mu, self.eta, self.beta, self.gamma]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 't_mat') is not None:
            self._parameters += ['-T', self.t_mat.tag]
        if getattr(self, 'my_mat') is not None:
            self._parameters += ['-My', self.my_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'orient_vals') is not None:
            self._parameters += ['-orient', *self.orient_vals]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        try:
            self.to_process(osi)
        except ValueError:
            self._parameters[0] = 'ElastomericBearingBoucWen'
            self.to_process(osi)



class FlatSliderBearing2D(ElementBase):
    """
    The FlatSliderBearing2D Element Class
    
    This command is used to construct a flatSliderBearing element object, which is defined by two nodes. The iNode
    represents the flat sliding surface and the jNode represents the slider. The element can have zero length or the
    appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D) friction properties for the
    shear deformations, and force-deformation behaviors defined by UniaxialMaterials in the remaining two (2D)
    or four (3D) directions. To capture the uplift behavior of the bearing, the user-specified
    UniaxialMaterial in the axial direction is modified for no-tension behavior. By default
    (sDratio = 0.0) P-Delta moments are entirely transferred to the flat sliding surface
    (iNode). It is important to note that rotations of the flat sliding surface
    (rotations at the iNode) affect the shear behavior of the bearing. To
    avoid the introduction of artificial viscous damping in the
    isolation system (sometimes referred to as "damping
    leakage in the isolation system"), the bearing
    element does not contribute to the Rayleigh
    damping by default. If the element has
    non-zero length, the local x-axis is
    determined from the nodal geometry
    unless the optional x-axis vector
    is specified in which case the
    nodal geometry is ignored and the user-defined orientation is utilized.

    For a two-dimensional problem
    """
    op_type = 'flatSliderBearing'

    def __init__(self, osi, ele_nodes, frn_mdl, k_init, p_mat=None, mz_mat=None, do_rayleigh=False, max_iter: int=None, tol: float=None, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for FlatSliderBearing2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        frn_mdl: obj
            Object associated with previously-defined frictionmodel
        k_init: float
            Initial elastic stiffness in local shear direction
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        max_iter: int, optional
            Maximum number of iterations to undertake to satisfy element equilibrium (optional, default = 20)
        tol: float, optional
            Convergence tolerance to satisfy element equilibrium (optional, default = 1e-8)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> o3.element.FlatSliderBearing2D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, k_init=1.0, p_mat=p_mat, mz_mat=mz_mat,
        >>>                                do_rayleigh=False, max_iter=1, tol=1.0, orient=None, mass=1.0, shear_dist=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.frn_mdl = frn_mdl
        self.k_init = float(k_init)
        self.p_mat = p_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        if max_iter is None:
            self.max_iter = None
        else:
            self.max_iter = int(max_iter)
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.frn_mdl.tag, self.k_init]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class FlatSliderBearing3D(ElementBase):
    """
    The FlatSliderBearing3D Element Class
    
    This command is used to construct a flatSliderBearing element object, which is defined by two nodes. The iNode
    represents the flat sliding surface and the jNode represents the slider. The element can have zero length or the
    appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D) friction properties for the
    shear deformations, and force-deformation behaviors defined by UniaxialMaterials in the remaining two (2D)
    or four (3D) directions. To capture the uplift behavior of the bearing, the user-specified
    UniaxialMaterial in the axial direction is modified for no-tension behavior. By default
    (sDratio = 0.0) P-Delta moments are entirely transferred to the flat sliding surface
    (iNode). It is important to note that rotations of the flat sliding surface
    (rotations at the iNode) affect the shear behavior of the bearing. To
    avoid the introduction of artificial viscous damping in the
    isolation system (sometimes referred to as "damping
    leakage in the isolation system"), the bearing
    element does not contribute to the Rayleigh
    damping by default. If the element has
    non-zero length, the local x-axis is
    determined from the nodal geometry
    unless the optional x-axis vector
    is specified in which case the
    nodal geometry is ignored and the user-defined orientation is utilized.

    For a three-dimensional problem
    """
    op_type = 'flatSliderBearing'

    def __init__(self, osi, ele_nodes, frn_mdl, k_init, p_mat=None, t_mat=None, my_mat=None, mz_mat=None, do_rayleigh=False, max_iter=None, tol: float=None, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for FlatSliderBearing3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        frn_mdl: obj
            Object associated with previously-defined frictionmodel
        k_init: float
            Initial elastic stiffness in local shear direction
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        t_mat: obj, optional
            Object associated with previously-defined uniaxial_material in torsional direction
        my_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local y-axis
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        max_iter: None, optional
            
        tol: float, optional
            Convergence tolerance to satisfy element equilibrium (optional, default = 1e-8)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [0, 1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> orient_vals = [1, 0, 0]
        >>> o3.element.FlatSliderBearing3D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, k_init=1.0, p_mat=p_mat, t_mat=t_mat,
        >>>                                my_mat=my_mat, mz_mat=mz_mat, do_rayleigh=False, max_iter=None, tol=None,
        >>>                                mass=1.0, shear_dist=1.0, orient=orient_vals)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.frn_mdl = frn_mdl
        self.k_init = float(k_init)
        self.p_mat = p_mat
        self.t_mat = t_mat
        self.my_mat = my_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        self.max_iter = max_iter
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.frn_mdl.tag, self.k_init]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 't_mat') is not None:
            self._parameters += ['-T', self.t_mat.tag]
        if getattr(self, 'my_mat') is not None:
            self._parameters += ['-My', self.my_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class SingleFPBearing2D(ElementBase):
    """
    The SingleFPBearing2D Element Class
    
    This command is used to construct a singleFPBearing element object, which is defined by two nodes. The iNode
    represents the concave sliding surface and the jNode represents the articulated slider. The element can have
    zero length or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D) friction
    properties (with post-yield stiffening due to the concave sliding surface) for the shear deformations, and
    force-deformation behaviors defined by UniaxialMaterials in the remaining two (2D) or four (3D)
    directions. To capture the uplift behavior of the bearing, the user-specified UniaxialMaterial
    in the axial direction is modified for no-tension behavior. By default (sDratio = 0.0)
    P-Delta moments are entirely transferred to the concave sliding surface (iNode). It
    is important to note that rotations of the concave sliding surface (rotations at
    the iNode) affect the shear behavior of the bearing. To avoid the introduction
    of artificial viscous damping in the isolation system (sometimes referred to
    as "damping leakage in the isolation system"), the bearing element does not
    contribute to the Rayleigh damping by default. If the element has non-zero
    length, the local x-axis is determined from the nodal geometry unless the
    optional x-axis vector is specified in which case the nodal geometry is
    ignored and the user-defined orientation is utilized.

    For a two-dimensional problem
    """
    op_type = 'singleFPBearing'

    def __init__(self, osi, ele_nodes, frn_mdl, reff, k_init, p_mat=None, mz_mat=None, do_rayleigh=False, max_iter: int=None, tol: float=None, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for SingleFPBearing2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        frn_mdl: obj
            Object associated with previously-defined frictionmodel
        reff: float
            Effective radius of concave sliding surface
        k_init: float
            Initial elastic stiffness in local shear direction
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        max_iter: int, optional
            Maximum number of iterations to undertake to satisfy element equilibrium (optional, default = 20)
        tol: float, optional
            Convergence tolerance to satisfy element equilibrium (optional, default = 1e-8)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> o3.element.SingleFPBearing2D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, reff=1.0, k_init=1.0, p_mat=p_mat,
        >>>                              mz_mat=mz_mat, do_rayleigh=False, max_iter=1, tol=1.0, orient=None,
        >>>                              mass=1.0, shear_dist=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.frn_mdl = frn_mdl
        self.reff = float(reff)
        self.k_init = float(k_init)
        self.p_mat = p_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        if max_iter is None:
            self.max_iter = None
        else:
            self.max_iter = int(max_iter)
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.frn_mdl.tag, self.reff, self.k_init]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class SingleFPBearing3D(ElementBase):
    """
    The SingleFPBearing3D Element Class
    
    This command is used to construct a singleFPBearing element object, which is defined by two nodes. The iNode
    represents the concave sliding surface and the jNode represents the articulated slider. The element can have
    zero length or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D) friction
    properties (with post-yield stiffening due to the concave sliding surface) for the shear deformations, and
    force-deformation behaviors defined by UniaxialMaterials in the remaining two (2D) or four (3D)
    directions. To capture the uplift behavior of the bearing, the user-specified UniaxialMaterial
    in the axial direction is modified for no-tension behavior. By default (sDratio = 0.0)
    P-Delta moments are entirely transferred to the concave sliding surface (iNode). It
    is important to note that rotations of the concave sliding surface (rotations at
    the iNode) affect the shear behavior of the bearing. To avoid the introduction
    of artificial viscous damping in the isolation system (sometimes referred to
    as "damping leakage in the isolation system"), the bearing element does not
    contribute to the Rayleigh damping by default. If the element has non-zero
    length, the local x-axis is determined from the nodal geometry unless the
    optional x-axis vector is specified in which case the nodal geometry is
    ignored and the user-defined orientation is utilized.

    For a three-dimensional problem
    """
    op_type = 'singleFPBearing'

    def __init__(self, osi, ele_nodes, frn_mdl, reff, k_init, p_mat=None, t_mat=None, my_mat=None, mz_mat=None, do_rayleigh=False, max_iter: int=None, tol: float=None, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for SingleFPBearing3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        frn_mdl: obj
            Object associated with previously-defined frictionmodel
        reff: float
            Effective radius of concave sliding surface
        k_init: float
            Initial elastic stiffness in local shear direction
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        t_mat: obj, optional
            Object associated with previously-defined uniaxial_material in torsional direction
        my_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local y axis
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        max_iter: int, optional
            Maximum number of iterations to undertake to satisfy element equilibrium (optional, default = 20)
        tol: float, optional
            Convergence tolerance to satisfy element equilibrium (optional, default = 1e-8)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [0, 1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> orient_vals = [1, 0, 0]
        >>> o3.element.SingleFPBearing3D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, reff=1.0, k_init=1.0, p_mat=p_mat, t_mat=t_mat,
        >>>                              my_mat=my_mat, mz_mat=mz_mat, do_rayleigh=False, max_iter=None, tol=None,
        >>>                              orient=orient_vals, mass=1.0, shear_dist=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.frn_mdl = frn_mdl
        self.reff = float(reff)
        self.k_init = float(k_init)
        self.p_mat = p_mat
        self.t_mat = t_mat
        self.my_mat = my_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        if max_iter is None:
            self.max_iter = None
        else:
            self.max_iter = int(max_iter)
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.frn_mdl.tag, self.reff, self.k_init]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 't_mat') is not None:
            self._parameters += ['-T', self.t_mat.tag]
        if getattr(self, 'my_mat') is not None:
            self._parameters += ['-My', self.my_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class TFP(ElementBase):
    """
    The TFP Element Class

    This command is used to construct a Triple Friction Pendulum Bearing element object, which is defined by two nodes.
    The element can have zero length or the appropriate bearing height. The bearing has unidirectional (2D) or coupled
    (3D) friction properties (with post-yield stiffening due to the concave sliding surface) for the shear
    deformations, and force-deformation behaviors defined by UniaxialMaterials in the remaining two (2D)
    or four (3D) directions. To capture the uplift behavior of the bearing, the user-specified
    UniaxialMaterial in the axial direction is modified for no-tension behavior. P-Delta
    moments are entirely transferred to the concave sliding surface (iNode). It is
    important to note that rotations of the concave sliding surface (rotations at
    the iNode) affect the shear behavior of the bearing. If the element has
    non-zero length, the local x-axis is determined from the nodal
    geometry unless the optional x-axis vector is specified in
    which case the nodal geometry is ignored and the user-defined orientation is utilized.


    """
    op_type = 'TFP'

    def __init__(self, osi, ele_nodes, r1, r2, r3, r4, db1, db2, db3, db4, d1, d2, d3, d4, mu1, mu2, mu3, mu4, h1, h2,
                 h3, h4, h0, col_load, big_k=None):
        """
        Initial method for TFP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        r1: float
            Radius of inner bottom sliding surface
        r2: float
            Radius of inner top sliding surface
        r3: float
            Radius of outer bottom sliding surface
        r4: float
            Radius of outer top sliding surface
        db1: float
            Diameter of inner bottom sliding surface
        db2: float
            Diameter of inner top sliding surface
        db3: float
            Diameter of outer bottom sliding surface
        db4: float
            Diameter of outer top sliding surface
        d1: float
            Diameter of inner slider
        d2: float
            Diameter of inner slider
        d3: float
            Diameter of outer bottom slider
        d4: float
            Diameter of outer top slider
        mu1: float
            Friction coefficient of inner bottom sliding surface
        mu2: float
            Friction coefficient of inner top sliding surface
        mu3: float
            Friction coefficient of outer bottom sliding surface
        mu4: float
            Friction coefficient of outer top sliding surface
        h1: float
            Height from inner bottom sliding surface to center of bearing
        h2: float
            Height from inner top sliding surface to center of bearing
        h3: float
            Height from outer bottom sliding surface to center of bearing
        h4: float
            Height from inner top sliding surface to center of bearing
        h0: float
            Total height of bearing
        col_load: float
            Initial axial load on bearing (only used for first time step then load come from model)
        big_k: float
            Optional, stiffness of spring in vertical dirn (dof 2 if ndm= 2, dof 3 if ndm = 3) (default=1.0e15)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.TFP(osi, ele_nodes=ele_nodes,
        >>>                r1=1.0, r2=1.0, r3=1.0, r4=1.0,
        >>>                db1=1.0, db2=1.0, db3=1.0, db4=1.0,
        >>>                d1=1.0, d2=1.0, d3=1.0, d4=1.0,
        >>>                mu1=0.3, mu2=0.4, mu3=0.5, mu4=0.5,
        >>>                h1=1.0, h2=1.0, h3=1.0, h4=1.0,
        >>>                h0=1.0, col_load=1.0, big_k=None)
        """
        self.osi = osi
        self.ele_nodes = [x.tag for x in ele_nodes]
        self.r1 = float(r1)
        self.r2 = float(r2)
        self.r3 = float(r3)
        self.r4 = float(r4)
        self.db1 = float(db1)
        self.db2 = float(db2)
        self.db3 = float(db3)
        self.db4 = float(db4)
        self.d1 = float(d1)
        self.d2 = float(d2)
        self.d3 = float(d3)
        self.d4 = float(d4)
        self.mu1 = float(mu1)
        self.mu2 = float(mu2)
        self.mu3 = float(mu3)
        self.mu4 = float(mu4)
        self.h1 = float(h1)
        self.h2 = float(h2)
        self.h3 = float(h3)
        self.h4 = float(h4)
        self.h0 = float(h0)
        self.col_load = float(col_load)
        if big_k is not None:
            self.big_k = float(big_k)
        else:
            self.big_k = None
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_nodes, self.r1, self.r2, self.r3, self.r4, self.db1,
                            self.db2, self.db3, self.db4, self.d1, self.d2, self.d3, self.d4, self.mu1, self.mu2,
                            self.mu3, self.mu4, self.h1, self.h2, self.h3, self.h4, self.h0, self.col_load]
        if getattr(self, 'big_k') is not None:
            self._parameters += [self.big_k]
        self.to_process(osi)



class TripleFrictionPendulum(ElementBase):
    """
    The TripleFrictionPendulum Element Class
    
    
    """
    op_type = 'TripleFrictionPendulum'

    def __init__(self, osi, ele_nodes, frn1, frn2, frn3, vert_mat, rot_z_mat, rot_x_mat, rot_y_mat, l1, l2, l3, d1, d2, d3, big_w, uy, kvt, min_fv, tol):
        """
        Initial method for TripleFrictionPendulum

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        frn1: obj
            = objects associated with previously-defined frictionmodels at the three sliding interfaces
        frn2: obj
            = objects associated with previously-defined frictionmodels at the three sliding interfaces
        frn3: obj
            = objects associated with previously-defined frictionmodels at the three sliding interfaces
        vert_mat: obj
            = pre-defined material object for compression behavior of the bearing
        rot_z_mat: obj
            = pre-defined material objects for rotational behavior about 3-axis, 1-axis and 2-axis, respectively.
        rot_x_mat: obj
            = pre-defined material objects for rotational behavior about 3-axis, 1-axis and 2-axis, respectively.
        rot_y_mat: obj
            = pre-defined material objects for rotational behavior about 3-axis, 1-axis and 2-axis, respectively.
        l1: float
            = effective radii. li = r_i - h_i (see figure 1)
        l2: float
            = effective radii. li = r_i - h_i (see figure 1)
        l3: float
            = effective radii. li = r_i - h_i (see figure 1)
        d1: float
            = displacement limits of pendulums (figure 1). displacement limit of the bearing is 2   ``d1`` +   ``d2`` + 
             ``d3`` +   ``l1``.   ``d3``/   ``l3`` -   ``l1``.   ``d2``/   ``l2``
        d2: float
            = displacement limits of pendulums (figure 1). displacement limit of the bearing is 2   ``d1`` +   ``d2`` + 
             ``d3`` +   ``l1``.   ``d3``/   ``l3`` -   ``l1``.   ``d2``/   ``l2``
        d3: float
            = displacement limits of pendulums (figure 1). displacement limit of the bearing is 2   ``d1`` +   ``d2`` + 
             ``d3`` +   ``l1``.   ``d3``/   ``l3`` -   ``l1``.   ``d2``/   ``l2``
        big_w: float
            = axial force used for the first trial of the first analysis step.
        uy: float
            = lateral displacement where sliding of the bearing starts. recommended value = 0.25 to 1 mm. a smaller
            value may cause convergence problem.
        kvt: float
            = tension stiffness k_vt of the bearing.
        min_fv: None
            
        tol: float
            = relative tolerance for checking the convergence of the element. recommended value = 1.e-10 to 1.e-3.

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> frn2 = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> frn3 = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> vert_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> rot_z_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> rot_x_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> rot_y_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> o3.element.TripleFrictionPendulum(osi, ele_nodes=ele_nodes, frn1=frn1, frn2=frn2, frn3=frn3, vert_mat=vert_mat,
        >>>                                   rot_z_mat=rot_z_mat, rot_x_mat=rot_x_mat, rot_y_mat=rot_y_mat, l1=1.0, l2=1.0,
        >>>                                   l3=1.0, d1=1.0, d2=1.0, d3=1.0, big_w=1.0, uy=1.0, kvt=1.0, min_fv=None, tol=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.frn1 = frn1
        self.frn2 = frn2
        self.frn3 = frn3
        self.vert_mat = vert_mat
        self.rot_z_mat = rot_z_mat
        self.rot_x_mat = rot_x_mat
        self.rot_y_mat = rot_y_mat
        self.l1 = float(l1)
        self.l2 = float(l2)
        self.l3 = float(l3)
        self.d1 = float(d1)
        self.d2 = float(d2)
        self.d3 = float(d3)
        self.big_w = float(big_w)
        self.uy = float(uy)
        self.kvt = float(kvt)
        self.min_fv = min_fv
        self.tol = float(tol)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.frn1.tag, self.frn2.tag, self.frn3.tag, self.vert_mat.tag, self.rot_z_mat.tag, self.rot_x_mat.tag, self.rot_y_mat.tag, self.l1, self.l2, self.l3, self.d1, self.d2, self.d3, self.big_w, self.uy, self.kvt, self.min_fv, self.tol]
        self.to_process(osi)


class MultipleShearSpring(ElementBase):
    """
    The MultipleShearSpring Element Class
    
    This command is used to construct a multipleShearSpring (MSS) element object, which is defined by two nodes. This
    element consists of a series of identical shear springs arranged radially to represent the isotropic behavior in the
    local y-z plane.

    
    """
    op_type = 'multipleShearSpring'

    def __init__(self, osi, ele_nodes, n_spring, mat=None, lim: float=None, mass: float=None, orient=None):
        """
        Initial method for MultipleShearSpring

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        n_spring: int
            Number of springs
        mat: obj, optional
            Object associated with previously-defined uniaxial_material object
        lim: float, optional
            Minimum deformation to calculate equivalent coefficient (see note 1)
        mass: float, optional
            Element mass
        orient: None, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [1, 0, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat = o3.uniaxial_material.Elastic(osi, 1.0)
        >>> o3.element.MultipleShearSpring(osi, ele_nodes=ele_nodes, n_spring=1, mat=mat, lim=1.0, mass=1.0, orient=None)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.n_spring = int(n_spring)
        self.mat = mat
        if lim is None:
            self.lim = None
        else:
            self.lim = float(lim)
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        self.orient = orient
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.n_spring]
        if getattr(self, 'mat') is not None:
            self._parameters += ['-mat', self.mat.tag]
        if getattr(self, 'lim') is not None:
            self._parameters += ['-lim', self.lim]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        self.to_process(osi)


class KikuchiBearingadjustPDOutput(ElementBase):
    """
    The KikuchiBearingadjustPDOutput Element Class
    
    This command is used to construct a KikuchiBearing element object, which is defined by two nodes. This element
    consists of multiple shear spring model (MSS) and multiple normal spring model (MNS).

    
    """
    op_type = 'KikuchiBearing'

    def __init__(self, osi, ele_nodes, total_rubber, ci, cj, shape: float=None, size: float=None, total_height: float=None, n_mss: int=None, mat_mss=None, lim_disp: float=None, n_mns: int=None, mat_mns=None, lamb: float=None, no_pd_input=False, no_tilt=False, orient=None, mass: float=None):
        """
        Initial method for KikuchiBearingadjustPDOutput

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        total_rubber: float
            Total rubber thickness
        ci: float
            P-delta moment adjustment for reaction force (default:    ``ci`` =0.5,    ``cj`` =0.5)
        cj: float
            P-delta moment adjustment for reaction force (default:    ``ci`` =0.5,    ``cj`` =0.5)
        shape: float, optional
            Following shapes are available: round, square
        size: float, optional
            Diameter (round shape), length of edge (square shape)
        total_height: float, optional
            Total height of the bearing (defaulut: distance between inode and jnode)
        n_mss: int, optional
            Number of springs in mss = nmss
        mat_mss: obj, optional
            Matobject for mss
        lim_disp: float, optional
            Minimum deformation to calculate equivalent coefficient of mss (see note 1)
        n_mns: int, optional
            Number of springs in mns = nmns*nmns (for round and square shape)
        mat_mns: obj, optional
            Matobject for mns
        lamb: float, optional
            Parameter to calculate compression modulus distribution on mns (see note 2)
        no_pd_input: bool
            Not consider p-delta moment
        no_tilt: bool
            Not consider tilt of rigid link
        orient: None, optional
            
        mass: float, optional
            Element mass

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat_mss = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mat_mns = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> o3.element.KikuchiBearingadjustPDOutput(osi, ele_nodes=ele_nodes, shape=1.0, size=1.0, total_rubber=1.0, total_height=1.0, n_mss=1, mat_mss=mat_mss, lim_disp=1.0, n_mns=1, mat_mns=mat_mns, lamb=1.0, no_pd_input="string", no_tilt="string", ci=1.0, cj=1.0, orient=[0.0, 0.0], mass=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        if shape is None:
            self.shape = None
        else:
            self.shape = float(shape)
        if size is None:
            self.size = None
        else:
            self.size = float(size)
        self.total_rubber = float(total_rubber)
        if total_height is None:
            self.total_height = None
        else:
            self.total_height = float(total_height)
        if n_mss is None:
            self.n_mss = None
        else:
            self.n_mss = int(n_mss)
        self.mat_mss = mat_mss
        if lim_disp is None:
            self.lim_disp = None
        else:
            self.lim_disp = float(lim_disp)
        if n_mns is None:
            self.n_mns = None
        else:
            self.n_mns = int(n_mns)
        self.mat_mns = mat_mns
        if lamb is None:
            self.lamb = None
        else:
            self.lamb = float(lamb)
        self.no_pd_input = no_pd_input
        self.no_tilt = no_tilt
        self.ci = float(ci)
        self.cj = float(cj)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.total_rubber, '-adjustPDOutput', self.ci, self.cj]
        if getattr(self, 'shape') is not None:
            self._parameters += ['-shape', self.shape]
        if getattr(self, 'size') is not None:
            self._parameters += ['-size', self.size]
        if getattr(self, 'total_height') is not None:
            self._parameters += ['-totalHeight', self.total_height]
        if getattr(self, 'n_mss') is not None:
            self._parameters += ['-nMSS', self.n_mss]
        if getattr(self, 'mat_mss') is not None:
            self._parameters += ['-matMSS', self.mat_mss.tag]
        if getattr(self, 'lim_disp') is not None:
            self._parameters += ['-limDisp', self.lim_disp]
        if getattr(self, 'n_mns') is not None:
            self._parameters += ['-nMNS', self.n_mns]
        if getattr(self, 'mat_mns') is not None:
            self._parameters += ['-matMNS', self.mat_mns.tag]
        if getattr(self, 'lamb') is not None:
            self._parameters += ['-lambda', self.lamb]
        if getattr(self, 'no_pd_input'):
            self._parameters += ['-noPDInput']
        if getattr(self, 'no_tilt'):
            self._parameters += ['-noTilt']
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)

class KikuchiBearingdoBalance(ElementBase):
    """
    The KikuchiBearingdoBalance Element Class
    
    This command is used to construct a KikuchiBearing element object, which is defined by two nodes. This element
    consists of multiple shear spring model (MSS) and multiple normal spring model (MNS).

    
    """
    op_type = 'KikuchiBearing'

    def __init__(self, osi, ele_nodes, total_rubber, lim_fo, lim_fi, n_iter, shape: float=None, size: float=None, total_height: float=None, n_mss: int=None, mat_mss=None, lim_disp: float=None, n_mns: int=None, mat_mns=None, lamb: float=None, no_pd_input=False, no_tilt=False, orient=None, mass: float=None):
        """
        Initial method for KikuchiBearingdoBalance

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        total_rubber: float
            Total rubber thickness
        lim_fo: float
            Tolerance of external unbalanced force (   ``limfo``), tolorance of internal unbalanced force (  
            ``limfi``), number of iterations to get rid of internal unbalanced force (   ``niter``)
        lim_fi: float
            Tolerance of external unbalanced force (   ``limfo``), tolorance of internal unbalanced force (  
            ``limfi``), number of iterations to get rid of internal unbalanced force (   ``niter``)
        n_iter: float
            Tolerance of external unbalanced force (   ``limfo``), tolorance of internal unbalanced force (  
            ``limfi``), number of iterations to get rid of internal unbalanced force (   ``niter``)
        shape: float, optional
            Following shapes are available: round, square
        size: float, optional
            Diameter (round shape), length of edge (square shape)
        total_height: float, optional
            Total height of the bearing (defaulut: distance between inode and jnode)
        n_mss: int, optional
            Number of springs in mss = nmss
        mat_mss: obj, optional
            Matobject for mss
        lim_disp: float, optional
            Minimum deformation to calculate equivalent coefficient of mss (see note 1)
        n_mns: int, optional
            Number of springs in mns = nmns*nmns (for round and square shape)
        mat_mns: obj, optional
            Matobject for mns
        lamb: float, optional
            Parameter to calculate compression modulus distribution on mns (see note 2)
        no_pd_input: bool
            Not consider p-delta moment
        no_tilt: bool
            Not consider tilt of rigid link
        orient: None, optional
            
        mass: float, optional
            Element mass

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> mat_mss = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> mat_mns = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> o3.element.KikuchiBearingdoBalance(osi, ele_nodes=ele_nodes, shape=1.0, size=1.0, total_rubber=1.0, total_height=1.0, n_mss=1, mat_mss=mat_mss, lim_disp=1.0, n_mns=1, mat_mns=mat_mns, lamb=1.0, no_pd_input="string", no_tilt="string", lim_fo=1.0, lim_fi=1.0, n_iter=1.0, orient=[0.0, 0.0], mass=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        if shape is None:
            self.shape = None
        else:
            self.shape = float(shape)
        if size is None:
            self.size = None
        else:
            self.size = float(size)
        self.total_rubber = float(total_rubber)
        if total_height is None:
            self.total_height = None
        else:
            self.total_height = float(total_height)
        if n_mss is None:
            self.n_mss = None
        else:
            self.n_mss = int(n_mss)
        self.mat_mss = mat_mss
        if lim_disp is None:
            self.lim_disp = None
        else:
            self.lim_disp = float(lim_disp)
        if n_mns is None:
            self.n_mns = None
        else:
            self.n_mns = int(n_mns)
        self.mat_mns = mat_mns
        if lamb is None:
            self.lamb = None
        else:
            self.lamb = float(lamb)
        self.no_pd_input = no_pd_input
        self.no_tilt = no_tilt
        self.lim_fo = float(lim_fo)
        self.lim_fi = float(lim_fi)
        self.n_iter = float(n_iter)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.total_rubber, '-doBalance', self.lim_fo, self.lim_fi, self.n_iter]
        if getattr(self, 'shape') is not None:
            self._parameters += ['-shape', self.shape]
        if getattr(self, 'size') is not None:
            self._parameters += ['-size', self.size]
        if getattr(self, 'total_height') is not None:
            self._parameters += ['-totalHeight', self.total_height]
        if getattr(self, 'n_mss') is not None:
            self._parameters += ['-nMSS', self.n_mss]
        if getattr(self, 'mat_mss') is not None:
            self._parameters += ['-matMSS', self.mat_mss.tag]
        if getattr(self, 'lim_disp') is not None:
            self._parameters += ['-limDisp', self.lim_disp]
        if getattr(self, 'n_mns') is not None:
            self._parameters += ['-nMNS', self.n_mns]
        if getattr(self, 'mat_mns') is not None:
            self._parameters += ['-matMNS', self.mat_mns.tag]
        if getattr(self, 'lamb') is not None:
            self._parameters += ['-lambda', self.lamb]
        if getattr(self, 'no_pd_input'):
            self._parameters += ['-noPDInput']
        if getattr(self, 'no_tilt'):
            self._parameters += ['-noTilt']
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)


class YamamotoBiaxialHDRcoRS(ElementBase):
    """
    The YamamotoBiaxialHDRcoRS Element Class
    
    This command is used to construct a YamamotoBiaxialHDR element object, which is defined by two nodes. This element
    can be used to represent the isotropic behavior of high-damping rubber bearing in the local y-z plane.

    
    """
    op_type = 'YamamotoBiaxialHDR'

    def __init__(self, osi, ele_nodes, tp, d_do, d_di, hr, cr, cs, orient: list=None, mass: float=None):
        """
        Initial method for YamamotoBiaxialHDRcoRS

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        tp: int
            Compound type = 1 : x0.6r manufactured by bridgestone corporation.
        d_do: float
            Outer diameter [m]
        d_di: float
            Bore diameter [m]
        hr: float
            Total thickness of rubber layer [m] optional data
        cr: float
            Coefficients for shear stress components of tau_r and tau_s
        cs: float
            Coefficients for shear stress components of tau_r and tau_s
        orient: list, optional
            
        mass: float, optional
            Element mass [kg]

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.YamamotoBiaxialHDRcoRS(osi, ele_nodes=ele_nodes, tp=1, d_do=1.0, d_di=1.0, hr=1.0, cr=1.0, cs=1.0, orient=[0.0, 0.0], mass=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.tp = int(tp)
        self.d_do = float(d_do)
        self.d_di = float(d_di)
        self.hr = float(hr)
        self.cr = float(cr)
        self.cs = float(cs)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.tp, self.d_do, self.d_di, self.hr, '-coRS', self.cr, self.cs]
        if getattr(self, 'orient') is not None:
            self._parameters += ['--orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)


class ElastomericX(ElementBase):
    """
    The ElastomericX Element Class
    
    This command is used to construct an ElastomericX bearing element object in three-dimension. The 3D continuum
    geometry of an elastomeric bearing is modeled as a 2-node, 12 DOF discrete element. This elements extends the
    formulation of Elastomeric_Bearing_(Bouc-Wen)_Element element. However, instead of the user providing
    material models as input arguments, it only requires geometric and material properties of an
    elastomeric bearing as arguments. The material models in six direction are formulated
    within the element from input arguments. The time-dependent values of mechanical
    properties (e.g., shear stiffness, buckling load capacity) can also be recorded
    using the "parameters" recorder.

    For 3D problem
    """
    op_type = 'ElastomericX'

    def __init__(self, osi, ele_nodes, fy, alpha, gr, kbulk, d1, d2, ts, tr, n, x1, x2, x3, y1, y2, y3, kc, phi_m, ac, s_dratio, m, cd, tc, tag1, tag2, tag3, tag4):
        """
        Initial method for ElastomericX

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        fy: float
            Yield strength
        alpha: float
            Post-yield stiffness ratio
        gr: float
            Shear modulus of elastomeric bearing
        kbulk: float
            Bulk modulus of rubber
        d1: float
            Internal diameter
        d2: float
            Outer diameter (excluding cover thickness)
        ts: float
            Single steel shim layer thickness
        tr: float
            Single  rubber layer thickness
        n: int
            Number of rubber layers
        x1: float
            Vector components in global coordinates defining local x-axis 
        x2: float
            Vector components in global coordinates defining local x-axis 
        x3: float
            Vector components in global coordinates defining local x-axis 
        y1: float
            Vector components in global coordinates defining local y-axis 
        y2: float
            Vector components in global coordinates defining local y-axis 
        y3: float
            Vector components in global coordinates defining local y-axis 
        kc: float
            Cavitation parameter (optional, default = 10.0)
        phi_m: float
            Damage parameter (optional, default = 0.5)
        ac: float
            Strength reduction parameter (optional, default = 1.0)
        s_dratio: float
            Shear distance from inode as a fraction of the element length (optional, default = 0.5)
        m: float
            Element mass (optional, default = 0.0)
        cd: float
            Viscous damping parameter (optional, default = 0.0)
        tc: float
            Cover thickness (optional, default = 0.0)
        tag1: float
            Object to include cavitation and post-cavitation (optional, default = 0)
        tag2: float
            Object to include buckling load variation (optional, default = 0)
        tag3: float
            Object to include horizontal stiffness variation (optional, default = 0)
        tag4: float
            Object to include vertical stiffness variation (optional, default = 0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.ElastomericX(osi, ele_nodes=ele_nodes, fy=1.0, alpha=1.0, gr=1.0, kbulk=1.0, d1=1.0, d2=1.0, ts=1.0,
        >>>                         tr=1.0, n=1, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, kc=1.0, phi_m=1.0, ac=1.0,
        >>>                         s_dratio=1.0, m=1.0, cd=1.0, tc=1.0, tag1=0, tag2=0, tag3=0, tag4=0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.fy = float(fy)
        self.alpha = float(alpha)
        self.gr = float(gr)
        self.kbulk = float(kbulk)
        self.d1 = float(d1)
        self.d2 = float(d2)
        self.ts = float(ts)
        self.tr = float(tr)
        self.n = int(n)
        self.x1 = float(x1)
        self.x2 = float(x2)
        self.x3 = float(x3)
        self.y1 = float(y1)
        self.y2 = float(y2)
        self.y3 = float(y3)
        self.kc = float(kc)
        self.phi_m = float(phi_m)
        self.ac = float(ac)
        self.s_dratio = float(s_dratio)
        self.m = float(m)
        self.cd = float(cd)
        self.tc = float(tc)
        self.tag1 = float(tag1)
        self.tag2 = float(tag2)
        self.tag3 = float(tag3)
        self.tag4 = float(tag4)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.fy, self.alpha, self.gr, self.kbulk, self.d1, self.d2, self.ts, self.tr, self.n, self.x1, self.x2, self.x3, self.y1, self.y2, self.y3, self.kc, self.phi_m, self.ac, self.s_dratio, self.m, self.cd, self.tc, self.tag1, self.tag2, self.tag3, self.tag4]
        self.to_process(osi)


class LeadRubberX(ElementBase):
    """
    The LeadRubberX Element Class
    
    This command is used to construct a LeadRubberX bearing element object in three-dimension. The 3D continuum geometry
    of a lead rubber bearing is modeled as a 2-node, 12 DOF discrete element. It extends the formulation of ElastomericX by
    including strength degradation in lead rubber bearing due to heating of the lead-core. The LeadRubberX element
    requires only the geometric and material properties of an elastomeric bearing as arguments. The material
    models in six direction are formulated within the element from input arguments. The time-dependent
    values of mechanical properties (e.g., shear stiffness, buckling load capacity, temperature in
    the lead-core, yield strength) can also be recorded using the "parameters" recorder.

    
    """
    op_type = 'LeadRubberX'

    def __init__(self, osi, ele_nodes, fy, alpha, gr, kbulk, d1, d2, ts, tr, n, x1, x2, x3, y1, y2, y3, kc, phi_m, ac, s_dratio, m, cd, tc, q_l, c_l, k_s, a_s, tag1, tag2, tag3, tag4, tag5):
        """
        Initial method for LeadRubberX

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        fy: float
            Yield strength
        alpha: float
            Post-yield stiffness ratio
        gr: float
            Shear modulus of elastomeric bearing
        kbulk: float
            Bulk modulus of rubber
        d1: float
            Internal diameter
        d2: float
            Outer diameter (excluding cover thickness)
        ts: float
            Single steel shim layer thickness
        tr: float
            Single rubber layer thickness
        n: int
            Number of rubber layers
        x1: float
            Vector components in global coordinates defining local x-axis 
        x2: float
            Vector components in global coordinates defining local x-axis 
        x3: float
            Vector components in global coordinates defining local x-axis 
        y1: float
            Vector components in global coordinates defining local y-axis 
        y2: float
            Vector components in global coordinates defining local y-axis 
        y3: float
            Vector components in global coordinates defining local y-axis 
        kc: float
            Cavitation parameter (optional, default = 10.0)
        phi_m: float
            Damage parameter (optional, default = 0.5)
        ac: float
            Strength reduction parameter (optional, default = 1.0)
        s_dratio: float
            Shear distance from inode as a fraction of the element length (optional, default = 0.5)
        m: float
            Element mass (optional, default = 0.0)
        cd: float
            Viscous damping parameter (optional, default = 0.0)
        tc: float
            Cover thickness (optional, default = 0.0)
        q_l: float
            Density of lead (optional, default = 11200 kg/m3)
        c_l: float
            Specific heat of lead (optional, default = 130 n-m/kg oc)
        k_s: float
            Thermal conductivity of steel (optional, default = 50 w/m oc)
        a_s: float
            Thermal diffusivity of steel (optional, default = 1.41e-05 m2/s)
        tag1: int
            Object to include cavitation and post-cavitation (optional, default = 0)
        tag2: int
            Object to include buckling load variation (optional, default = 0)
        tag3: int
            Object to include horizontal stiffness variation (optional, default = 0)
        tag4: int
            Object to include vertical stiffness variation (optional, default = 0)
        tag5: int
            Object to include strength degradation in shear due to heating of lead core (optional, default = 0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.LeadRubberX(osi, ele_nodes=ele_nodes, fy=1.0, alpha=1.0, gr=1.0, kbulk=1.0, d1=1.0, d2=1.0, ts=1.0,
        >>>                        tr=1.0, n=1, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, kc=1.0, phi_m=1.0, ac=1.0,
        >>>                        s_dratio=1.0, m=1.0, cd=1.0, tc=1.0, q_l=1.0, c_l=1.0, k_s=1.0, a_s=1.0,
        >>>                        tag1=1, tag2=1, tag3=1, tag4=1, tag5=1)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.fy = float(fy)
        self.alpha = float(alpha)
        self.gr = float(gr)
        self.kbulk = float(kbulk)
        self.d1 = float(d1)
        self.d2 = float(d2)
        self.ts = float(ts)
        self.tr = float(tr)
        self.n = int(n)
        self.x1 = float(x1)
        self.x2 = float(x2)
        self.x3 = float(x3)
        self.y1 = float(y1)
        self.y2 = float(y2)
        self.y3 = float(y3)
        self.kc = float(kc)
        self.phi_m = float(phi_m)
        self.ac = float(ac)
        self.s_dratio = float(s_dratio)
        self.m = float(m)
        self.cd = float(cd)
        self.tc = float(tc)
        self.q_l = float(q_l)
        self.c_l = float(c_l)
        self.k_s = float(k_s)
        self.a_s = float(a_s)
        self.tag1 = int(tag1)
        self.tag2 = int(tag2)
        self.tag3 = int(tag3)
        self.tag4 = int(tag4)
        self.tag5 = int(tag5)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.fy, self.alpha, self.gr, self.kbulk, self.d1, self.d2, self.ts, self.tr, self.n, self.x1, self.x2, self.x3, self.y1, self.y2, self.y3, self.kc, self.phi_m, self.ac, self.s_dratio, self.m, self.cd, self.tc, self.q_l, self.c_l, self.k_s, self.a_s, self.tag1, self.tag2, self.tag3, self.tag4, self.tag5]
        self.to_process(osi)


class HDR(ElementBase):
    """
    The HDR Element Class
    
    For 3D problem
    """
    op_type = 'HDR'

    def __init__(self, osi, ele_nodes, gr, kbulk, d1, d2, ts, tr, n, a1, a2, a3, b1, b2, b3, c1, c2, c3, c4, x1, x2, x3, y1, y2, y3, kc, phi_m, ac, s_dratio, m, tc):
        """
        Initial method for HDR

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        gr: float
            Shear modulus of elastomeric bearing
        kbulk: float
            Bulk modulus of rubber
        d1: float
            Internal diameter
        d2: float
            Outer diameter (excluding cover thickness)
        ts: float
            Single steel shim layer thickness
        tr: float
            Single rubber layer thickness
        n: int
            Number of rubber layers
        a1: float
            Parameters of the grant model
        a2: float
            Parameters of the grant model
        a3: float
            Parameters of the grant model
        b1: float
            Parameters of the grant model
        b2: float
            Parameters of the grant model
        b3: float
            Parameters of the grant model
        c1: float
            Parameters of the grant model
        c2: float
            Parameters of the grant model
        c3: float
            Parameters of the grant model
        c4: float
            Parameters of the grant model
        x1: float
            Vector components in global coordinates defining local x-axis 
        x2: float
            Vector components in global coordinates defining local x-axis 
        x3: float
            Vector components in global coordinates defining local x-axis 
        y1: float
            Vector components in global coordinates defining local y-axis 
        y2: float
            Vector components in global coordinates defining local y-axis 
        y3: float
            Vector components in global coordinates defining local y-axis 
        kc: float
            Cavitation parameter (optional, default = 10.0)
        phi_m: float
            Damage parameter (optional, default = 0.5)
        ac: float
            Strength reduction parameter (optional, default = 1.0)
        s_dratio: float
            Shear distance from inode as a fraction of the element length (optional, default = 0.5)
        m: float
            Element mass (optional, default = 0.0)
        tc: float
            Cover thickness (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.HDR(osi, ele_nodes=ele_nodes, gr=1.0, kbulk=1.0, d1=1.0, d2=1.0, ts=1.0, tr=1.0, n=1, a1=1.0, a2=1.0, a3=1.0, b1=1.0, b2=1.0, b3=1.0, c1=1.0, c2=1.0, c3=1.0, c4=1.0, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, kc=1.0, phi_m=1.0, ac=1.0, s_dratio=1.0, m=1.0, tc=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.gr = float(gr)
        self.kbulk = float(kbulk)
        self.d1 = float(d1)
        self.d2 = float(d2)
        self.ts = float(ts)
        self.tr = float(tr)
        self.n = int(n)
        self.a1 = float(a1)
        self.a2 = float(a2)
        self.a3 = float(a3)
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.b3 = float(b3)
        self.c1 = float(c1)
        self.c2 = float(c2)
        self.c3 = float(c3)
        self.c4 = float(c4)
        self.x1 = float(x1)
        self.x2 = float(x2)
        self.x3 = float(x3)
        self.y1 = float(y1)
        self.y2 = float(y2)
        self.y3 = float(y3)
        self.kc = float(kc)
        self.phi_m = float(phi_m)
        self.ac = float(ac)
        self.s_dratio = float(s_dratio)
        self.m = float(m)
        self.tc = float(tc)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.gr, self.kbulk, self.d1, self.d2, self.ts, self.tr, self.n, self.a1, self.a2, self.a3, self.b1, self.b2, self.b3, self.c1, self.c2, self.c3, self.c4, self.x1, self.x2, self.x3, self.y1, self.y2, self.y3, self.kc, self.phi_m, self.ac, self.s_dratio, self.m, self.tc]
        self.to_process(osi)


class RJWatsonEqsBearing2D(ElementBase):
    """
    The RJWatsonEqsBearing2D Element Class
    
    This command is used to construct a RJWatsonEqsBearing element object, which is defined by two nodes. The iNode
    represents the masonry plate and the jNode represents the sliding surface plate. The element can have zero length
    or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D) friction properties (with
    post-yield stiffening due to the mass-energy-regulator (MER) springs) for the shear deformations, and
    force-deformation behaviors defined by UniaxialMaterials in the remaining two (2D) or four (3D)
    directions. To capture the uplift behavior of the bearing, the user-specified UniaxialMaterial
    in the axial direction is modified for no-tension behavior. By default (sDratio = 1.0)
    P-Delta moments are entirely transferred to the sliding surface (jNode). It is
    important to note that rotations of the sliding surface (rotations at the
    jNode) affect the shear behavior of the bearing. To avoid the
    introduction of artificial viscous damping in the isolation
    system (sometimes referred to as "damping leakage in the
    isolation system"), the bearing element does not
    contribute to the Rayleigh damping by default.
    If the element has non-zero length, the local
    x-axis is determined from the nodal geometry
    unless the optional x-axis vector is
    specified in which case the nodal
    geometry is ignored and the user-defined orientation is utilized.

    For a two-dimensional problem
    """
    op_type = 'RJWatsonEqsBearing'

    def __init__(self, osi, ele_nodes, frn_mdl, k_init, p_mat=None, vy_mat=None, mz_mat=None, do_rayleigh=False, max_iter: int=None, tol: float=None, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for RJWatsonEqsBearing2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        frn_mdl: obj
            Object associated with previously-defined frictionmodel
        k_init: float
            Initial stiffness of sliding friction component in local shear direction
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        vy_mat: obj, optional
            Object associated with previously-defined uniaxial_material in shear direction along local y-axis (mer
            spring behavior not including friction)
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        max_iter: int, optional
            Maximum number of iterations to undertake to satisfy element equilibrium (optional, default = 20)
        tol: float, optional
            Convergence tolerance to satisfy element equilibrium (optional, default = 1e-8)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> vy_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> frn_mdl = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> o3.element.RJWatsonEqsBearing2D(osi, ele_nodes=ele_nodes, frn_mdl=frn_mdl, k_init=1.0, p_mat=p_mat, vy_mat=vy_mat,
        >>>                                 mz_mat=mz_mat, do_rayleigh=False, max_iter=1, tol=1.0, orient=None, mass=1.0,
        >>>                                 shear_dist=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.frn_mdl = frn_mdl
        self.k_init = float(k_init)
        self.p_mat = p_mat
        self.vy_mat = vy_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        if max_iter is None:
            self.max_iter = None
        else:
            self.max_iter = int(max_iter)
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.frn_mdl.tag, self.k_init]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 'vy_mat') is not None:
            self._parameters += ['-Vy', self.vy_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class RJWatsonEqsBearing3D(ElementBase):
    """
    The RJWatsonEqsBearing3D Element Class
    
    This command is used to construct a RJWatsonEqsBearing element object, which is defined by two nodes. The iNode
    represents the masonry plate and the jNode represents the sliding surface plate. The element can have zero length
    or the appropriate bearing height. The bearing has unidirectional (2D) or coupled (3D) friction properties (with
    post-yield stiffening due to the mass-energy-regulator (MER) springs) for the shear deformations, and
    force-deformation behaviors defined by UniaxialMaterials in the remaining two (2D) or four (3D)
    directions. To capture the uplift behavior of the bearing, the user-specified UniaxialMaterial
    in the axial direction is modified for no-tension behavior. By default (sDratio = 1.0)
    P-Delta moments are entirely transferred to the sliding surface (jNode). It is
    important to note that rotations of the sliding surface (rotations at the
    jNode) affect the shear behavior of the bearing. To avoid the
    introduction of artificial viscous damping in the isolation
    system (sometimes referred to as "damping leakage in the
    isolation system"), the bearing element does not
    contribute to the Rayleigh damping by default.
    If the element has non-zero length, the local
    x-axis is determined from the nodal geometry
    unless the optional x-axis vector is
    specified in which case the nodal
    geometry is ignored and the user-defined orientation is utilized.

    For a three-dimensional problem
    """
    op_type = 'RJWatsonEqsBearing'

    def __init__(self, osi, ele_nodes, frn_mdl, k_init, p_mat=None, vy_mat=None, vz_mat=None, t_mat=None, my_mat=None, mz_mat=None, do_rayleigh=False, max_iter: int=None, tol: float=None, orient=None, mass: float=None, shear_dist: float=None):
        """
        Initial method for RJWatsonEqsBearing3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        frn_mdl: obj
            Object associated with previously-defined frictionmodel
        k_init: float
            Initial stiffness of sliding friction component in local shear direction
        p_mat: obj, optional
            Object associated with previously-defined uniaxial_material in axial direction
        vy_mat: obj, optional
            Object associated with previously-defined uniaxial_material in shear direction along local y-axis (mer
            spring behavior not including friction)
        vz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in shear direction along local z-axis (mer
            spring behavior not including friction)
        t_mat: obj, optional
            Object associated with previously-defined uniaxial_material in torsional direction
        my_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local y-axis
        mz_mat: obj, optional
            Object associated with previously-defined uniaxial_material in moment direction around local z-axis
        do_rayleigh: bool
            To include rayleigh damping from the bearing (optional, default = no rayleigh damping contribution)
        max_iter: int, optional
            Maximum number of iterations to undertake to satisfy element equilibrium (optional, default = 20)
        tol: float, optional
            Convergence tolerance to satisfy element equilibrium (optional, default = 1e-8)
        orient: None, optional
            
        mass: float, optional
            Element mass (optional, default = 0.0)
        shear_dist: float, optional
            Shear distance from inode as a fraction of the element length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3, ndf=6)
        >>> coords = [[0, 0, 0], [0, 1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> p_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> vy_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> vz_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> t_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> my_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> mz_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
        >>> orient_vals = [1, 0, 0]
        >>> frn_mdl = o3.friction_model.Coulomb(osi, mu=1.0)
        >>> o3.element.RJWatsonEqsBearing3D(osi, ele_nodes=ele_nodes, frn_mdl=frn_mdl, k_init=1.0, p_mat=p_mat,
        >>>                                 vy_mat=vy_mat, vz_mat=vz_mat, t_mat=t_mat, my_mat=my_mat, mz_mat=mz_mat,
        >>>                                 do_rayleigh=False, max_iter=1, tol=1.0, orient=orient_vals, mass=1.0, shear_dist=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.frn_mdl = frn_mdl
        self.k_init = float(k_init)
        self.p_mat = p_mat
        self.vy_mat = vy_mat
        self.vz_mat = vz_mat
        self.t_mat = t_mat
        self.my_mat = my_mat
        self.mz_mat = mz_mat
        self.do_rayleigh = do_rayleigh
        if max_iter is None:
            self.max_iter = None
        else:
            self.max_iter = int(max_iter)
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        self.orient = orient
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        if shear_dist is None:
            self.shear_dist = None
        else:
            self.shear_dist = float(shear_dist)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.frn_mdl.tag, self.k_init]
        if getattr(self, 'p_mat') is not None:
            self._parameters += ['-P', self.p_mat.tag]
        if getattr(self, 'vy_mat') is not None:
            self._parameters += ['-Vy', self.vy_mat.tag]
        if getattr(self, 'vz_mat') is not None:
            self._parameters += ['-Vz', self.vz_mat.tag]
        if getattr(self, 't_mat') is not None:
            self._parameters += ['-T', self.t_mat.tag]
        if getattr(self, 'my_mat') is not None:
            self._parameters += ['-My', self.my_mat.tag]
        if getattr(self, 'mz_mat') is not None:
            self._parameters += ['-Mz', self.mz_mat.tag]
        if getattr(self, 'do_rayleigh'):
            self._parameters += ['-doRayleigh']
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'orient') is not None:
            self._parameters += ['-orient', *self.orient]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'shear_dist') is not None:
            self._parameters += ['-shearDist', self.shear_dist]
        self.to_process(osi)


class FPBearingPTV(ElementBase):
    """
    The FPBearingPTV Element Class
    
    The FPBearingPTV command creates a single Friction Pendulum bearing element, which is capable of accounting for the
    changes in the coefficient of friction at the sliding surface with instantaneous values of the sliding velocity, axial
    pressure and temperature at the sliding surface. The constitutive modelling is similar to the existing
    singleFPBearing element, otherwise. The FPBearingPTV element has been verified and validated in
    accordance with the ASME guidelines, details of which are presented in Chapter 4 of Kumar et al. (2015a).

    
    """
    op_type = 'FPBearingPTV'

    def __init__(self, osi, ele_nodes, mu_ref, is_pressure_dependent, p_ref, is_temperature_dependent, diffusivity, conductivity, is_velocity_dependent, rate_parameter, reffective_fp, radius__contact, k_initial, the_material_a, the_material_b, the_material_c, the_material_d, x1, x2, x3, y1, y2, y3, shear_dist, do_rayleigh, mass, max_iter, tol, unit):
        """
        Initial method for FPBearingPTV

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        mu_ref: float
            Reference coefficient of friction
        is_pressure_dependent: int
            1 if the coefficient of friction is a function of instantaneous axial pressure
        p_ref: float
            Reference axial pressure (the bearing pressure under static loads)
        is_temperature_dependent: int
            1 if the coefficient of friction is a function of instantaneous temperature at the sliding surface
        diffusivity: float
            Thermal diffusivity of steel
        conductivity: float
            Thermal conductivity of steel
        is_velocity_dependent: int
            1 if the coefficient of friction is a function of instantaneous velocity at the sliding surface
        rate_parameter: float
            The exponent that determines the shape of the coefficient of friction vs. sliding velocity curve
        reffective_fp: float
            Effective radius of curvature of the sliding surface of the fpbearing
        radius__contact: float
            Radius of contact area at the sliding surface
        k_initial: float
            Lateral  stiffness of the sliding bearing before sliding begins
        the_material_a: int
            Object for the uniaxial material in the axial direction
        the_material_b: int
            Object for the uniaxial material in the torsional direction
        the_material_c: int
            Object for the uniaxial material for rocking about local y axis
        the_material_d: int
            Object for the uniaxial material for rocking about local z axis
        x1: float
            Vector components to define local x axis
        x2: float
            Vector components to define local x axis
        x3: float
            Vector components to define local x axis
        y1: float
            Vector components to define local y axis
        y2: float
            Vector components to define local y axis
        y3: float
            Vector components to define local y axis
        shear_dist: float
            Shear distance from inode as a fraction of the length of the element
        do_rayleigh: int
            To include rayleigh damping from the bearing
        mass: float
            Element mass
        max_iter: int
            Maximum number of iterations to satisfy the equilibrium of element
        tol: float
            Convergence tolerance to satisfy the equilibrium of the element
        unit: int
            Object to identify the unit from the list below. * ``1``: n, m, s, c * ``2``: kn, m, s, c * ``3``: n, mm, s,
            c * ``4``: kn, mm, s, c * ``5``: lb, in, s, c * ``6``: kip, in, s, c * ``7``: lb, ft, s, c * ``8``: kip, ft, s, c

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.FPBearingPTV(osi, ele_nodes=ele_nodes, mu_ref=1.0, is_pressure_dependent=1, p_ref=1.0, is_temperature_dependent=1, diffusivity=1.0, conductivity=1.0, is_velocity_dependent=1, rate_parameter=1.0, reffective_fp=1.0, radius__contact=1.0, k_initial=1.0, the_material_a=1, the_material_b=1, the_material_c=1, the_material_d=1, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, shear_dist=1.0, do_rayleigh=1, mass=1.0, max_iter=1, tol=1.0, unit=1)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mu_ref = float(mu_ref)
        self.is_pressure_dependent = int(is_pressure_dependent)
        self.p_ref = float(p_ref)
        self.is_temperature_dependent = int(is_temperature_dependent)
        self.diffusivity = float(diffusivity)
        self.conductivity = float(conductivity)
        self.is_velocity_dependent = int(is_velocity_dependent)
        self.rate_parameter = float(rate_parameter)
        self.reffective_fp = float(reffective_fp)
        self.radius__contact = float(radius__contact)
        self.k_initial = float(k_initial)
        self.the_material_a = int(the_material_a)
        self.the_material_b = int(the_material_b)
        self.the_material_c = int(the_material_c)
        self.the_material_d = int(the_material_d)
        self.x1 = float(x1)
        self.x2 = float(x2)
        self.x3 = float(x3)
        self.y1 = float(y1)
        self.y2 = float(y2)
        self.y3 = float(y3)
        self.shear_dist = float(shear_dist)
        self.do_rayleigh = int(do_rayleigh)
        self.mass = float(mass)
        self.max_iter = int(max_iter)
        self.tol = float(tol)
        self.unit = int(unit)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mu_ref, self.is_pressure_dependent, self.p_ref, self.is_temperature_dependent, self.diffusivity, self.conductivity, self.is_velocity_dependent, self.rate_parameter, self.reffective_fp, self.radius__contact, self.k_initial, self.the_material_a, self.the_material_b, self.the_material_c, self.the_material_d, self.x1, self.x2, self.x3, self.y1, self.y2, self.y3, self.shear_dist, self.do_rayleigh, self.mass, self.max_iter, self.tol, self.unit]
        self.to_process(osi)



from o3seespy.command.nd_material.base_material import NDMaterialBase


class ContactMaterial2D(NDMaterialBase):
    """
    The ContactMaterial2D NDMaterial Class
    
    This command is used to construct a ContactMaterial2D nDMaterial object.
    """
    op_type = 'ContactMaterial2D'

    def __init__(self, osi, mu, g_mod, c, t):
        """
        Initial method for ContactMaterial2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mu: float
            Interface frictional coefficient
        g_mod: float
            Interface stiffness parameter
        c: float
            Interface cohesive intercept
        t: float
            Interface tensile strength

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.ContactMaterial2D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
        """
        self.osi = osi
        self.mu = float(mu)
        self.g_mod = float(g_mod)
        self.c = float(c)
        self.t = float(t)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.mu, self.g_mod, self.c, self.t]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class ContactMaterial3D(NDMaterialBase):
    """
    The ContactMaterial3D NDMaterial Class
    
    This command is used to construct a ContactMaterial3D nDMaterial object.
    """
    op_type = 'ContactMaterial3D'

    def __init__(self, osi, mu, g_mod, c, t):
        """
        Initial method for ContactMaterial3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mu: float
            Interface frictional coefficient
        g_mod: float
            Interface stiffness parameter
        c: float
            Interface cohesive intercept
        t: float
            Interface tensile strength

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> o3.nd_material.ContactMaterial3D(osi, mu=1.0, g_mod=1.0, c=1.0, t=1.0)
        """
        self.osi = osi
        self.mu = float(mu)
        self.g_mod = float(g_mod)
        self.c = float(c)
        self.t = float(t)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.mu, self.g_mod, self.c, self.t]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



import o3seespy as o3
import numpy as np


def test_cantilever_w_force_beam_column():
    def run_w_settings(ele_len, e_mod, i_sect, pload, udl, beam_in_parts, use_pload):
        osi = o3.OpenSeesInstance(ndm=2, state=3)

        # Establish nodes
        left_node = o3.node.Node(osi, 0, 0)
        right_node = o3.node.Node(osi, ele_len, 0)

        # Fix bottom node
        o3.Fix3DOF(osi, left_node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
        o3.Fix3DOF(osi, right_node, o3.cc.FREE, o3.cc.FREE, o3.cc.FREE)

        area = 0.5
        lp_i = 0.2
        lp_j = 0.2

        if beam_in_parts:
            elastic = 1

            if elastic:
                left_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
            else:
                m_cap = 600.0
                b = 0.05
                phi = m_cap / (e_mod * i_sect)
                # mat_flex = o3.uniaxial_material.Steel01(osi, m_cap, e0=e_mod * i_sect, b=b)
                mat_flex = o3.uniaxial_material.ElasticBilin(osi, e_mod * i_sect, e_mod * i_sect * b, phi)
                mat_axial = o3.uniaxial_material.Elastic(osi, e_mod * area)
                left_sect = o3.section.Aggregator(osi, mats=[mat_axial.tag, o3.cc.P, mat_flex.tag, o3.cc.M_Z, mat_flex.tag,
                                                             o3.cc.M_Y])
            right_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
            centre_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
            integ = o3.beam_integration.HingeMidpoint(osi, left_sect, lp_i, right_sect, lp_j, centre_sect)
            beam_transf = o3.geom_transf.Linear2D(osi, )
            ele = o3.element.ForceBeamColumn(osi, [left_node, right_node], beam_transf, integ)
        else:
            beam_transf = o3.geom_transf.Linear2D(osi)
            ele = o3.element.ElasticBeamColumn2D(osi, [left_node, right_node], area, e_mod, i_sect, beam_transf)
        # Apply gravity loads

        # If true then load applied along beam

        ts_po = o3.time_series.Linear(osi, factor=1)
        o3.pattern.Plain(osi, ts_po)
        if use_pload:
            o3.Load(osi, right_node, [0, -pload, 0])
        else:
            o3.EleLoad2DUniform(osi, ele, -udl)

        tol = 1.0e-3
        o3.constraints.Plain(osi)
        o3.numberer.RCM(osi)
        o3.system.BandGen(osi)
        n_steps_gravity = 10
        o3.integrator.LoadControl(osi, 1. / n_steps_gravity, num_iter=10)
        o3.test_check.NormDispIncr(osi, tol, 10)
        o3.algorithm.Linear(osi)
        o3.analysis.Static(osi)
        o3.analyze(osi, n_steps_gravity)
        o3.gen_reactions(osi)
        end_disp = o3.get_node_disp(osi, right_node, dof=o3.cc.DOF_Y)
        return o3.get_ele_response(osi, ele, 'force')[:3], end_disp

    ele_len_o = 4.0
    pload_o = 10.0 * ele_len_o
    udl_o = 10.0
    e_mod_o = 200.0
    i_sect_o = 0.1

    # point load
    w_pload = 1
    disp_expected = -pload_o * ele_len_o ** 3 / (3 * e_mod_o * i_sect_o)
    r, disp = run_w_settings(ele_len_o, e_mod_o, i_sect_o, pload_o, udl_o, beam_in_parts=0, use_pload=w_pload)
    assert np.isclose(disp, disp_expected, rtol=0.01)
    assert np.isclose(r[1], pload_o, rtol=0.01)
    assert np.isclose(r[2], pload_o * ele_len_o, rtol=0.01)
    r, disp = run_w_settings(ele_len_o, e_mod_o, i_sect_o, pload_o, udl_o, beam_in_parts=1, use_pload=w_pload)
    assert np.isclose(disp, disp_expected, rtol=0.01), (disp, disp_expected)
    assert np.isclose(r[1], pload_o, rtol=0.01)
    assert np.isclose(r[2], pload_o * ele_len_o, rtol=0.01)

    # UDL
    w_pload = 0
    disp_expected = -udl_o * ele_len_o ** 4 / (8 * e_mod_o * i_sect_o)
    v_expected = udl_o * ele_len_o
    m_expected = udl_o * ele_len_o ** 2 / 2
    r, disp = run_w_settings(ele_len_o, e_mod_o, i_sect_o, pload_o, udl_o, beam_in_parts=0, use_pload=w_pload)
    assert np.isclose(disp, disp_expected, rtol=0.01), (disp, disp_expected)
    assert np.isclose(r[1], v_expected, rtol=0.01)
    assert np.isclose(r[2], m_expected, rtol=0.01)
    r, disp = run_w_settings(ele_len_o, e_mod_o, i_sect_o, pload_o, udl_o, beam_in_parts=1, use_pload=w_pload)
    assert np.isclose(disp, disp_expected, rtol=0.05), (disp, disp_expected)
    assert np.isclose(r[1], v_expected, rtol=0.01)
    assert np.isclose(r[2], m_expected, rtol=0.02), (r[2], m_expected)


def test_disp_control_cantilever_nonlinear():
    osi = o3.OpenSeesInstance(ndm=2, state=3)

    ele_len = 4.0

    # Establish nodes
    left_node = o3.node.Node(osi, 0, 0)
    right_node = o3.node.Node(osi, ele_len, 0)

    # Fix bottom node
    o3.Fix3DOF(osi, left_node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, right_node, o3.cc.FREE, o3.cc.FREE, o3.cc.FREE)
    e_mod = 200.0
    i_sect = 0.1
    area = 0.5
    lp_i = 0.1
    lp_j = 0.1

    m_cap = 0.30
    b = 0.05
    phi = m_cap / (e_mod * i_sect)
    # mat_flex = o3.uniaxial_material.Steel01(osi, m_cap, e0=e_mod * i_sect, b=b)
    mat_flex = o3.uniaxial_material.ElasticBilin(osi, e_mod * i_sect, e_mod * i_sect * b, phi)
    mat_axial = o3.uniaxial_material.Elastic(osi, e_mod * area)
    left_sect = o3.section.Aggregator(osi, mats=[[mat_axial, o3.cc.P], [mat_flex, o3.cc.M_Z], [mat_flex, o3.cc.M_Y]])
    right_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
    centre_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
    integ = o3.beam_integration.HingeMidpoint(osi, left_sect, lp_i, right_sect, lp_j, centre_sect)
    beam_transf = o3.geom_transf.Linear2D(osi, )
    ele = o3.element.ForceBeamColumn(osi, [left_node, right_node], beam_transf, integ)

    # start displacement controlled
    d_inc = 0.01

    # opy.wipeAnalysis()
    o3.constraints.Plain(osi)
    o3.numberer.RCM(osi)
    o3.system.BandGen(osi)
    o3.test_check.NormUnbalance(osi, 2, max_iter=10, p_flag=0)
    # o3.test_check.FixedNumIter(osi, max_iter=10)
    # o3.test_check.NormDispIncr(osi, 0.002, 10, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.integrator.DisplacementControl(osi, right_node, o3.cc.DOF_Y, d_inc)
    o3.analysis.Static(osi)
    ts_po = o3.time_series.Linear(osi, factor=1)
    o3.pattern.Plain(osi, ts_po)
    o3.Load(osi, right_node, [0.0, 1.0, 0])
    o3.analyze(osi, 4)
    end_disp = o3.get_node_disp(osi, right_node, dof=o3.cc.DOF_Y)
    r = o3.get_ele_response(osi, ele, 'force')[:3]
    k = r[1] / -end_disp
    k_elastic_expected = 1. / (ele_len ** 3 / (3 * e_mod * i_sect))
    assert np.isclose(k, k_elastic_expected)
    o3.analyze(osi, 6)
    end_disp = o3.get_node_disp(osi, right_node, dof=o3.cc.DOF_Y)
    r = o3.get_ele_response(osi, ele, 'force')[:3]
    k = r[1] / -end_disp
    assert k < 0.95 * k_elastic_expected


def run():
    osi = o3.OpenSeesInstance(ndm=2, state=3)

    ele_len = 4.0

    # Establish nodes
    left_node = o3.node.Node(osi, 0, 0)
    right_node = o3.node.Node(osi, ele_len, 0)

    # Fix bottom node
    o3.Fix(osi, left_node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix(osi, right_node, o3.cc.FREE, o3.cc.FREE, o3.cc.FREE)
    e_mod = 200.0e2
    i_sect = 0.1
    area = 0.5
    lp_i = 0.1
    lp_j = 0.1
    beam_in_parts = 1
    if beam_in_parts:
        elastic = 0

        if elastic:
            left_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
        else:
            m_cap = 18.80
            b = 0.05
            phi = m_cap / (e_mod * i_sect)
            # mat_flex = o3.uniaxial_material.Steel01(osi, m_cap, e0=e_mod * i_sect, b=b)
            mat_flex = o3.uniaxial_material.ElasticBilin(osi, e_mod * i_sect, e_mod * i_sect * b, phi)
            mat_axial = o3.uniaxial_material.Elastic(osi, e_mod * area)
            left_sect = o3.section.Aggregator(osi, mats=[mat_axial.tag, o3.cc.P, mat_flex.tag, o3.cc.M_Z, mat_flex.tag, o3.cc.M_Y])
        right_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
        centre_sect = o3.section.Elastic2D(osi, e_mod, area, i_sect)
        integ = o3.beam_integration.HingeMidpoint(osi, left_sect, lp_i, right_sect, lp_j, centre_sect)
        beam_transf = o3.geom_transf.Linear2D(osi, )
        ele = o3.element.ForceBeamColumn(osi, [left_node, right_node], beam_transf, integ)
    else:
        beam_transf = o3.geom_transf.Linear2D(osi)
        ele = o3.element.ElasticBeamColumn2D(osi, [left_node, right_node], area, e_mod, i_sect, beam_transf)

    w_gloads = 1
    if w_gloads:
        # Apply gravity loads
        # If true then load applied along beam
        use_pload = 0
        pload = 1.0 * ele_len
        udl = 2.0
        ts_po = o3.time_series.Linear(osi, factor=1)
        o3.pattern.Plain(osi, ts_po)
        if use_pload:
            o3.Load(osi, right_node, [0, -pload, 0])
        else:
            o3.EleLoad2DUniform(osi, ele, -udl)

        tol = 1.0e-3
        o3.constraints.Plain(osi)
        o3.numberer.RCM(osi)
        o3.system.BandGeneral(osi)
        n_steps_gravity = 10
        o3.integrator.LoadControl(osi, 1. / n_steps_gravity, num_iter=10)
        o3.test_check.NormDispIncr(osi, tol, 10)
        o3.algorithm.Linear(osi)
        o3.analysis.Static(osi)
        o3.analyze(osi, n_steps_gravity)
        o3.gen_reactions(osi)
        print('reactions: ', o3.get_ele_response(osi, ele, 'force')[:3])
        end_disp = o3.get_node_disp(osi, right_node, dof=o3.cc.Y)
        print(f'end_disp: {end_disp}')
        if use_pload:
            disp_expected = pload * ele_len ** 3 / (3 * e_mod * i_sect)
            print(f'v_expected: {pload}')
            print(f'm_expected: {pload * ele_len}')
            print(f'disp_expected: {disp_expected}')
        else:
            v_expected = udl * ele_len
            m_expected = udl * ele_len ** 2 / 2
            disp_expected = udl * ele_len ** 4 / (8 * e_mod * i_sect)
            print(f'v_expected: {v_expected}')
            print(f'm_expected: {m_expected}')
            print(f'disp_expected: {disp_expected}')

        # o3.extensions.to_py_file(osi, 'temp4.py')
    disp_load = 1
    if disp_load:
        o3.load_constant(osi, time=0.0)
        end_disp_init = o3.get_node_disp(osi, right_node, dof=o3.cc.Y)
        # start displacement controlled
        d_inc = -0.01

        # opy.wipeAnalysis()
        o3.numberer.RCM(osi)
        o3.system.BandGeneral(osi)
        o3.test_check.NormUnbalance(osi, 2, max_iter=10, p_flag=0)
        # o3.test_check.FixedNumIter(osi, max_iter=10)
        # o3.test_check.NormDispIncr(osi, 0.002, 10, p_flag=0)
        o3.algorithm.Newton(osi)
        o3.integrator.DisplacementControl(osi, right_node, o3.cc.Y, d_inc)
        o3.analysis.Static(osi)
        ts_po = o3.time_series.Linear(osi, factor=1)
        o3.pattern.Plain(osi, ts_po)
        o3.Load(osi, right_node, [0.0, 1.0, 0])
        ok = o3.analyze(osi, 10)
        end_disp = o3.get_node_disp(osi, right_node, dof=o3.cc.Y)
        print(f'end_disp: {end_disp}')
        r = o3.get_ele_response(osi, ele, 'force')[:3]
        print('reactions: ', r)
        k = r[1] / -(end_disp - end_disp_init)
        print('k: ', k)
        k_elastic_expected = 1. / (ele_len ** 3 / (3 * e_mod * i_sect))
        print('k_elastic_expected: ', k_elastic_expected)


if __name__ == '__main__':
    # test_disp_control_cantilever_nonlinear()
    run()








def to_commands(op_base_type, parameters):
    para = []
    for i, e in enumerate(parameters):
        if isinstance(e, str):
            e = "'%s'" % e
        elif isinstance(e, float):
            e = '%.6g' % e
            if '.' not in e and 'e' not in e:
                e += '.0'
        para.append(str(e))
        if i > 40:  # avoid verbose print output
            break
    p_str = ', '.join(para)
    return 'opy.%s(%s)' % (op_base_type, p_str)

#
# def _dep_get_fn_name_and_args(line):
#     import re
#     line = line.replace('.', 'stoppoint')
#     fn_match = re.match(r"(?P<function>\w+)\s?\((?P<arg>(?P<args>\w+(,\s?)?)+)\)", line)
#     fn_dict = fn_match.groupdict()
#     args = [arg.strip() for arg in fn_dict['arg'].split(',')]
#     args = [arg.replace('stoppoint', '.') for arg in args]
#     return fn_dict['function'], args


def _get_fn_name_and_args(line):
    import re
    fn_name = line.split('(')[0]
    args = re.search(r'\((.*?)\)', line).group(1)
    args = args.replace(' ', '')
    args = args.split(',')
    for i in range(len(args)):

        if "'" in args[i] or '"' in args[i]:
            pass
        #     args[i] = args[i][1:-1]
        elif '.' in args[i]:
            args[i] = float(args[i])
        else:
            try:
                args[i] = int(args[i])
            except ValueError:
                pass

    return fn_name, args


def check_if_opy_lines_consistent(line1, line2, line3=None):
    if line3 is None:
        if line1 == line2:
            return True
        fn1, args1 = _get_fn_name_and_args(line1)
        fn2, args2 = _get_fn_name_and_args(line2)
        if fn1 == fn2 and len(args1) == len(args2):
            for i in range(len(args1)):
                if args1[i] == args2[i]:
                    continue
                elif isinstance(args1[i], str) or isinstance(args2, str):
                    # consistent = 0  # since strings must be identical
                    return False
            return True
        else:
            return False
    else:
        if line1 == line2 and line1 == line3:
            return True
        fn1, args1 = _get_fn_name_and_args(line1)
        fn2, args2 = _get_fn_name_and_args(line2)
        fn3, args3 = _get_fn_name_and_args(line3)
        if fn1 == fn2 and fn1 == fn3 and len(args1) == len(args2) and len(args1) == len(args3):
            for i in range(len(args1)):
                if args1[i] == args2[i] and args1[i] == args3[i]:
                    continue
                elif isinstance(args1[i], str) or isinstance(args2[i], str) or isinstance(args3[i], str):
                    # consistent = 0  # since strings must be identical
                    return False
                else:
                    val1 = float(args1[i])
                    val2 = float(args2[i])
                    val3 = float(args3[i])
                    if val2 - val1 == val3 - val2:
                        continue
                    else:
                        return False
            return True
        else:
            return False


def _build_logic_formula(line1, line2):
    if line1 == line2:
        return line1
    fn1, args1 = _get_fn_name_and_args(line1)
    fn2, args2 = _get_fn_name_and_args(line2)
    new_args = list(args1)
    for i in range(len(args1)):
        if args1[i] == args2[i]:
            new_args[i] = str(args1[i])
        else:
            diff = args2[i] - args1[i]
            if diff < 0:
                new_args[i] = f'{args1[i]} {diff} * i'
            else:
                new_args[i] = f'{args1[i]} + {diff} * i'
    return f'{fn1}({", ".join(new_args)})'


def compress_opy_lines(commands):
    slines = 10  # search lines
    latest_rep = -1
    new_commands = []
    for i, com in enumerate(commands):
        if i < latest_rep:
            continue
        new_commands.append(com)
        dup_detected = 0
        for j in range(1, slines):
            if dup_detected or i + j >= len(commands) - 1:
                break
            if check_if_opy_lines_consistent(com, commands[i + j]):
                # check all lines in between are repeated
                consistent = 1
                for k in range(1, j):
                    if i + j + k >= len(commands):
                        consistent = 0
                        break
                    if not check_if_opy_lines_consistent(commands[i + k], commands[i + j + k]):
                        consistent = 0
                if consistent:
                    nr = 0
                    new_rep = i + nr * j
                    while consistent:
                        nr += 1
                        new_rep = i + nr * j
                        for k in range(j):
                            if new_rep + k >= len(commands):
                                consistent = 0
                            elif nr == 1:
                                if not check_if_opy_lines_consistent(commands[i + k], commands[new_rep + k]):
                                    consistent = 0
                                    break
                            else:
                                if not check_if_opy_lines_consistent(commands[i + (nr - 2) * j + k],
                                                                     commands[i + (nr - 1) * j + k],
                                                                     commands[i + nr * j + k]
                                                                     ):
                                    consistent = 0
                    if nr > 3:
                        dup_detected = 1
                        latest_rep = new_rep
                        new_commands[-1] = f'for i in range({nr}):'  # replace
                        for k in range(j):
                            new_command = _build_logic_formula(commands[i + k], commands[i + k + j])
                            new_commands.append('    ' + new_command)

    return new_commands


def to_py_file(osi, ofile='ofile.py', compress=True, w_analyze=False):
    if compress:
        commands = compress_opy_lines(osi.commands)
    else:
        commands = osi.commands
    pstr = 'import openseespy.opensees as opy\n' + '\n'.join(commands)
    if w_analyze:
        pstr += '\nopy.analyze(1, 0.1)\n'
    ofile = open(ofile, 'w')
    ofile.write(pstr)
    ofile.close()


def to_tcl_file(osi, ofile='ofile.tcl', w_analyze=False):
    # if compress:
    #     commands = compress_opy_lines(osi.commands)
    # else:
    commands = osi.commands
    pstr = '\n'.join(commands)
    if w_analyze:
        pstr += '\nopy.analyze(1, 0.1)\n'
    tcl_str = py2tcl(pstr)
    ofile = open(ofile, 'w')
    ofile.write(tcl_str)
    ofile.close()


def get_o3_kwargs_from_obj(obj, o3_obj, custom=None, overrides=None):
    from inspect import signature
    from collections import OrderedDict
    if custom is None:
        custom = {}
    if overrides is None:
        overrides = {}
    sig = signature(o3_obj)
    kwargs = OrderedDict()
    args = []
    sig_vals = sig.parameters.values()
    for p in sig_vals:
        if p.name in custom:
            pname = custom[p.name]
        else:
            pname = p.name
        if pname == 'osi':
            continue
        if pname in overrides:
            val = overrides[pname]
        else:
            try:
                val = getattr(obj, pname)
            except AttributeError as e:
                if p.default == p.empty:
                    raise AttributeError(e)
                else:
                    val = p.default
        if p.default == p.empty:
            args.append(val)
        else:
            if val is not None:
                kwargs[p.name] = val
    return args, kwargs


def has_o3_model_changed(cur_type, prev_type, cur_args, prev_args, cur_kwargs, prev_kwargs):
    import numpy as np
    changed = 0
    if cur_type != prev_type or len(cur_args) != len(prev_args) or len(cur_kwargs) != len(prev_kwargs):
        changed = 1
    else:
        for j, arg in enumerate(cur_args):
            if hasattr(arg, '__len__'):
                if len(arg) != len(prev_args[j]):
                    changed = 1
                    break
                for k, subarg in enumerate(arg):
                    if not np.isclose(subarg, prev_args[j][k]):
                        changed = 1
            elif not np.isclose(arg, prev_args[j]):
                changed = 1
                break
        for pm in cur_kwargs:
            if pm not in prev_kwargs:
                changed = 1
                break
            if hasattr(cur_kwargs[pm], '__len__'):
                if len(cur_kwargs[pm]) != len(prev_kwargs[pm]):
                    changed = 1
                    break
                for k, subarg in enumerate(cur_kwargs[pm]):
                    if not np.isclose(subarg, cur_kwargs[pm][k]):
                        changed = 1
                        break
            elif not np.isclose(cur_kwargs[pm], prev_kwargs[pm]):
                changed = 1
                break
    return changed


def py2tcl(pystr):
    """
    Converts openseespy script to tcl

    Returns
    -------

    """
    # new = '\n'.join(pystr.split()[1:])  # removes the import line
    new = pystr.replace('(', ' ')
    new = new.replace(')', ' ')
    new = new.replace('opy.', '')
    new = new.replace(',', '')
    new = new.replace("'", '')
    new = new.replace('"', '')
    # lines = new.splitlines()
    # for i in range(len(lines)):
    #     if 'for i in range(' in lines[i]:
    #         line = lines.replace('for i in range(', 'for {set i 1} {$i <= num} {incr i 1} {')
    return new


def gen_free_field_2d_bc(osi, eles, left_bc, bl_node=0, width=1, connection=None, base_fix_x=False, base_fix_y=True):
    import numpy as np
    from o3seespy import element
    from o3seespy import EqualDOF, Fix2DOFMulti
    from o3seespy import node
    # eles array_like of vertical quad like elements
    # bl_node is index of bottom-left node
    if left_bc:
        top_ind = 3
        bot_ind = 0
        sgn = -1
    else:
        top_ind = 2
        bot_ind = 1
        sgn = 1
    top_inds = np.array([3, 2])
    bot_inds = np.array([0, 1])  # TODO: depends on node order
    new_eles = []
    new_nodes = []
    for i, ele in enumerate(eles):
        if i == 0:
            top_line_nodes = np.array(ele.ele_nodes)[top_inds]
            new_nodes_line = []
            for nod in top_line_nodes:
                new_nodes_line.append(node.Node(osi, nod.x + sgn * width, nod.y))
            new_nodes.append(new_nodes_line)
            EqualDOF(osi, new_nodes[i][0], new_nodes[i][1], [1, 2])
            if connection is None:  # use rigid
                EqualDOF(osi, ele.ele_nodes[top_ind], new_nodes[i][0], [1, 2])
            # TODO: support uniaxial material based connections
        line_nodes = np.array(ele.ele_nodes)[bot_inds]

        new_nodes_line = []
        for nod in line_nodes:
            new_nodes_line.append(node.Node(osi, nod.x + sgn * width, nod.y))
        new_nodes.append(new_nodes_line)
        EqualDOF(osi, new_nodes[i][0], new_nodes[i][1], [1, 2])
        if connection is None:  # use rigid
            EqualDOF(osi, ele.ele_nodes[bot_ind], new_nodes[i][0], [1, 2])
        if isinstance(ele, element.SSPquad):
            ele_nodes = [new_nodes[i+1][0], new_nodes[i+1][1], new_nodes[i][1], new_nodes[i][0]]
            new_eles.append(element.SSPquad(osi, ele_nodes, ele.mat, ele.otype, ele.thick * 1e4,
                                            ele.b1, ele.b2))
        # TODO: support Quad
    if base_fix_x and base_fix_y:
        Fix2DOFMulti(osi, new_nodes[-1], [1, 2])
    elif base_fix_x:
        Fix2DOFMulti(osi, new_nodes[-1], x=1, y=0)
    else:
        Fix2DOFMulti(osi, new_nodes[-1], x=0, y=1)




from o3seespy.base_model import OpenSeesObject
from o3seespy.command.common import get_node_tags, get_ele_tags


class NodeRegion(OpenSeesObject):
    op_base_type = "region"
    op_type = None

    def __init__(self, osi, nodes, rayleigh=None):
        """
        A region defined by a group of nodes

        Parameters
        ----------
        osi : opensees_pack.opensees_instance.OpenSeesInstance object
            An instance of opensees
        nodes : str or list
            list of Nodes
        rayleigh: dict
            Dictionary of Rayleigh parameters
        """
        self.osi = osi
        osi.n_region += 1
        self._tag = osi.n_region
        self._parameters = [self._tag, "-node"]

        if isinstance(nodes, str) and nodes == 'all':
            self.nodes = 'all'
            self._parameters += get_node_tags(osi)
        else:
            self.nodes = [x.tag for x in nodes]
            self._parameters += self.nodes
        self.rayleigh = rayleigh
        if rayleigh is not None:
            pms = ['alpha_m', 'beta_k', 'beta_k_init', 'beta_k_comm']
            self._parameters.append('-rayleigh')
            for pm in pms:
                if pm in rayleigh:
                    self._parameters.append(rayleigh[pm])
                else:
                    self._parameters.append(0.0)
        self.to_process(osi)


class ElementRegion(OpenSeesObject):
    op_base_type = "region"
    op_type = None

    def __init__(self, osi, eles, rayleigh=None):
        """
        A region defined by a group of elements

        Parameters
        ----------
        osi : opensees_pack.opensees_instance.OpenSeesInstance object
            An instance of opensees
        eles : str or list
            list of elements
        rayleigh: dict
            Dictionary of Rayleigh parameters
        """
        self.osi = osi
        osi.n_region += 1
        self._tag = osi.n_region
        self._parameters = [self._tag, "-ele"]

        if isinstance(eles, str) and eles == 'all':
            self.eles = 'all'
            self._parameters += get_ele_tags(osi)
        else:
            self.eles = [x.tag for x in eles]
            self._parameters += self.eles
        self.rayleigh = rayleigh
        if rayleigh is not None:
            pms = ['alpha_m', 'beta_k', 'beta_k_init', 'beta_k_comm']
            self._parameters.append('-rayleigh')
            for pm in pms:
                if pm in rayleigh:
                    self._parameters.append(rayleigh[pm])
                else:
                    self._parameters.append(0.0)
        self.to_process(osi)



from o3seespy.command.element.base_element import ElementBase


class CatenaryCable(ElementBase):
    """
    The CatenaryCable Element Class
    
    This command is used to construct a catenary cable element object.

    
    """
    op_type = 'CatenaryCable'

    def __init__(self, osi, i_node, j_node, weight, big_e, big_a, l0, alpha, temperature_change, rho, error_tol, nsubsteps, mass_type):
        """
        Initial method for CatenaryCable

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        i_node: obj
            End nodes (3 dof per node)
        j_node: obj
            End nodes (3 dof per node)
        weight: float
            Undefined
        big_e: float
            Elastic modulus of the cable material
        big_a: float
            Cross-sectional area of element
        l0: float
            Unstretched length of the cable
        alpha: float
            Coefficient of thermal expansion
        temperature_change: float
            Temperature change for the element
        rho: float
            Mass per unit length
        error_tol: float
            Allowed tolerance for within-element equilbrium (newton-rhapson iterations)
        nsubsteps: int
            Number of within-element substeps into which equilibrium iterations are subdivided (not number of steps to
            convergence)
        mass_type: int
            Mass matrix model to use (``masstype`` = 0 lumped mass matrix,    ``masstype`` = 1 rigid-body mass matrix
            (in development))

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0)
        >>> o3.element.CatenaryCable(osi, i_node=i_node, j_node=j_node, weight=1.0, big_e=1.0, big_a=1.0, l0=1.0, alpha=1.0, temperature_change=1.0, rho=1.0, error_tol=1.0, nsubsteps=1, mass_type=1)
        """
        self.osi = osi
        self.i_node = i_node
        self.j_node = j_node
        self.weight = float(weight)
        self.big_e = float(big_e)
        self.big_a = float(big_a)
        self.l0 = float(l0)
        self.alpha = float(alpha)
        self.temperature_change = float(temperature_change)
        self.rho = float(rho)
        self.error_tol = float(error_tol)
        self.nsubsteps = int(nsubsteps)
        self.mass_type = int(mass_type)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.i_node.tag, self.j_node.tag, self.weight, self.big_e, self.big_a, self.l0, self.alpha, self.temperature_change, self.rho, self.error_tol, self.nsubsteps, self.mass_type]
        self.to_process(osi)



from o3seespy.command.element.base_element import ElementBase


class FourNodeTetrahedron(ElementBase):
    """
    The FourNodeTetrahedron Element Class
    
    This command is used to construct a standard four-node tetrahedron element objec with one-point Gauss integration.


       
    """
    op_type = 'FourNodeTetrahedron'

    def __init__(self, osi, ele_nodes, mat, b1, b2, b3):
        """
        Initial method for FourNodeTetrahedron

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes
        mat: obj
            Object of ndmaterial
        b1: float
            Body forces in global x,y,z directions
        b2: float
            Body forces in global x,y,z directions
        b3: float
            Body forces in global x,y,z directions

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> o3.element.FourNodeTetrahedron(osi, ele_nodes=ele_nodes, mat=mat, b1=1.0, b2=1.0, b3=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.b1 = float(b1)
        self.b2 = float(b2)
        self.b3 = float(b3)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.b1, self.b2, self.b3]
        self.to_process(osi)



import o3seespy as o3


def test_newton():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    o3.algorithm.Newton(osi)


def test_secant_newton():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    o3.algorithm.SecantNewton(osi)


def test_linear():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    o3.algorithm.Linear(osi)


def test_raphson_newton():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    o3.algorithm.RaphsonNewton(osi)


def test_krylov_newton():
    osi = o3.OpenSeesInstance(ndm=2, ndf=2)
    o3.test_check.NormDispIncr(osi, tol=1.0e-3, max_iter=35, p_flag=0)
    o3.algorithm.KrylovNewton(osi)


if __name__ == '__main__':
    test_krylov_newton()



import o3seespy as o3  # for testing only


def test_initial_state_analysis_wrapper():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
    o3.nd_material.InitialStateAnalysisWrapper(osi, n_d_mat=mat, n_dim=1)


def test_init_stress_nd_material():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
    o3.nd_material.InitStressNDMaterial(osi, other=mat, init_stress=1.0, n_dim=2)


# def test_init_strain_nd_material():
#     osi = o3.OpenSeesInstance(ndm=2)
#     mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
#     o3.nd_material.InitStrainNDMaterial(osi, other=mat, init_strain=1.0, n_dim=2)


from o3seespy.command.element.base_element import *
from .beam_column import *
from .bearing import *
from .brick import *
from .cable import *
from .contact import *
from .joint import *
from .link import *
from .misc import *
from .other_up import *
from .pfem import *
from .quadrilateral import *
from .tetrahedron import *
from .triangular import *
from .truss import *
from .uc_san_diego_up import *
from .zero_length import *
